import numpy as np

from ..broker.interfaces import IOrderSide
from .base_alpha import BaseAlpha
from ..insight.insight import Insight, StrategyDependantConfirmation


class EMAPriceCrossoverAlpha(BaseAlpha):
    """
    ### EMA Crossover Alpha
    This alpha model generates insights based on EMA Crossover.

    :param strategy (BaseStrategy): The strategy instance
    :param atrPeriod (int): The period for the ATR indicator
    :param emaPeriod (int): The period for the EMA indicator
    :param baseConfidenceModifierField (str): The field to use for modifying the base confidence

    Author:
        @isaac-diaby
    """
    atrColumn: str
    emaColumn: str

    def __init__(self, strategy, atrPeriod: int = 14, emaPeriod: int = 14, baseConfidenceModifierField = None):
        super().__init__(strategy, "EMA_PRICE_CROSSOVER", "0.1", baseConfidenceModifierField)
        # Check if atrPeriod and emaPeriod are above 0
        if atrPeriod <= 0 or emaPeriod <= 0:
            raise ValueError("ATR Period and EMA Period must be positive integers")

        self.TA = [
            {"kind": 'atr', "length": atrPeriod},
            {"kind": 'ema', "length": emaPeriod}
        ]
        self.STRATEGY.warm_up = max(atrPeriod, emaPeriod)
        self.atrColumn = f'ATRr_{atrPeriod}'
        self.emaColumn = f'EMA_{emaPeriod}'

    def start(self):
        pass

    def init(self, asset):
        pass

    def generateInsights(self, symbol):
        try:
            latestBar = self.get_latest_bar(symbol)
            previousBar = self.get_previos_bar(symbol)
            latestIATR = latestBar[self.atrColumn]
            latestBarIEMA = latestBar[self.emaColumn]
            previousBarIATR = previousBar[self.atrColumn]
            previousBarIEMA = previousBar[self.emaColumn]

            baseConfidence = self.STRATEGY.baseConfidence
            # Modify Confidence based on baseConfidenceModifierField
            if (self.baseConfidenceModifierField):
                baseConfidence *= abs(self.get_baseConfidenceModifier(symbol))

            if baseConfidence <= 0:
                return self.returnResults(message="Base Confidence is 0.")

            if self.is_long_signal(latestBar, previousBar, latestIATR, latestBarIEMA, previousBarIATR, previousBarIEMA):
                return self.create_insight(IOrderSide.BUY, symbol, latestBar, previousBar, latestIATR, latestBarIEMA, baseConfidence)

            if self.is_short_signal(latestBar, previousBar, latestIATR, latestBarIEMA, previousBarIATR, previousBarIEMA, symbol):
                return self.create_insight(IOrderSide.SELL, symbol, latestBar, previousBar, latestIATR, latestBarIEMA, baseConfidence)

            return self.returnResults()
        except Exception as e:
            return self.returnResults(success=False, message=str(e))

    def is_long_signal(self, latestBar, previousBar, latestIATR, latestBarIEMA, previousBarIATR, previousBarIEMA):
        return (latestBarIEMA < latestBar['close'] and previousBarIEMA > previousBar['high'] and
                # np.abs(previousBar['open'] - previousBar['close']) > previousBarIATR and
                np.abs(latestBar['close'] - latestBarIEMA) < latestIATR)

    def is_short_signal(self, latestBar, previousBar, latestIATR, latestBarIEMA, previousBarIATR, previousBarIEMA, symbol):
        return (self.STRATEGY.assets[symbol]['shortable'] and latestBarIEMA > latestBar['close'] and
                previousBarIEMA < previousBar['low'] and
                # np.abs(previousBar['open'] - previousBar['close']) > previousBarIATR and
                np.abs(latestBarIEMA - latestBar['close']) < latestIATR)

    def create_insight(self, order_side: IOrderSide, symbol: str, latestBar, previousBar, latestIATR, latestBarIEMA, baseConfidence: float):
        if order_side == IOrderSide.BUY:
            TP = self.STRATEGY.tools.dynamic_round(latestBar['high']+(latestIATR*3.5), symbol)
            SL = self.STRATEGY.tools.dynamic_round(max(previousBar['low']-(.5*latestIATR), latestBarIEMA-latestIATR*1.5), symbol)
        else:
            TP = self.STRATEGY.tools.dynamic_round(latestBar['low']-(latestIATR*3.5), symbol)
            SL = self.STRATEGY.tools.dynamic_round(min(previousBar['high']+(.5*latestIATR), latestBarIEMA+latestIATR*1.5), symbol)

        ENTRY = self.STRATEGY.tools.dynamic_round(latestBarIEMA, symbol)
        TTLUF = self.STRATEGY.tools.calculateTimeToLive(latestBar['close'], ENTRY, latestIATR)
        TTLF = self.STRATEGY.tools.calculateTimeToLive(TP, ENTRY, latestIATR)

        return self.returnResults(Insight(order_side, symbol, self.NAME, self.STRATEGY.resolution, None, ENTRY, [TP], SL, baseConfidence, [StrategyDependantConfirmation.NONE], TTLUF, TTLF))



from datetime import datetime
from dash import dcc, html, register_page, callback, no_update
from dash.dependencies import Input, Output, State
from dash_tvlwc import Tvlwc
from dash_tvlwc.types import SeriesType, ColorType
import plotly.graph_objects as go


from OlympusTrader.broker.interfaces import IAccount, IPosition
from OlympusTrader.insight.insight import Insight
from OlympusTrader.strategy.interfaces import IStrategyMatrics, IStrategyMode
from OlympusTrader.ui.components.accountBalanceCard import AccountBalanceCard
from OlympusTrader.ui.components.dashboardCards import dashboardCard, dashboardCardInsights, dashboardCardMode
from OlympusTrader.ui.components.matricItem import MatricItem
from OlympusTrader.ui.components.placeholderChart import PlaceholderChart
from OlympusTrader.ui.components.dashboardNav import subnav
from OlympusTrader.ui.components.tradesTable import tradeTable
from OlympusTrader.ui.interfaces.store import STRATEGY_STORE_MAPPINGS

register_page(__name__, title="OlympusTrader - Overview", path='/')

# TESTING
# balance_series =[{"time": time.strftime('%Y-%m-%d'), "value": value} for time, value in zip(
#     pd.date_range(start="2023-01-01", periods=100), np.random.randint(40_000, 60_000, 100))]


def account_card_section_populate(account: IAccount, metrics: IStrategyMatrics, insights: Insight, mode: IStrategyMode):
    alltimeChange = round(account["equity"] - metrics["starting_cash"], 2)

    # print(insight)
    # print(metrics)
    return [AccountBalanceCard(balance=round(account["equity"], 2), alltimeChange=alltimeChange, dailyChange=alltimeChange),
            dashboardCardInsights(title="Insights", insights=insights),
            # dashboardCard(title="Win Rate", value=f"{metrics["win_rate"]*100}%", icon="fa-thin fa-chart-pie"),
            dashboardCardMode(title="Mode", mode=mode, accountType=account['account_id'], connected=True)
            ]


account_card_section = html.Div(
    id="account-card-section", className="grid grid-cols-1 md:grid-cols-2 lg:grid-cols-3 gap-6")


@callback(
    Output("account-card-section", "children"),
    [
        Input(STRATEGY_STORE_MAPPINGS.account.id, 'data'),
        Input(STRATEGY_STORE_MAPPINGS.metrics.id, 'data'),
        Input(STRATEGY_STORE_MAPPINGS.insights.id, 'data'),
        Input(STRATEGY_STORE_MAPPINGS.mode.id, 'data'),

    ],
    # prevent_initial_call=True,

)
def update_account_card_section(account, metrics, insights, mode):

    return account_card_section_populate(account, metrics, insights, mode)

strategy_metrics_section = html.Div(
        className="bg-primary-foreground rounded-lg shadow-lg border border-accent p-6 mt-8 ",
        children=[
            html.H3("Strategy Metrics",
                    className="text-xl font-semibold text-accent mb-4"),
            html.Div(
                id="strategy-metrics",
                className="grid grid-cols-2 md:grid-cols-3 lg:grid-cols-4 gap-4",
                children=[
                    # MatricItem(label="Win Rate", value=assets[idx]['winrate']),
                    # MatricItem(label="Profit Factor",
                    #            value=assets[idx]['profit_factor']),
                    # MatricItem(label="Sharpe Ratio",
                    #            value=assets[idx]['sharpe_ratio']),
                    # MatricItem(label="Max Drawdown",
                    #            value=assets[idx]['max_drawdown']),
                    # MatricItem(label="Avg Win Amount",
                    #            value=assets[idx]['avg_win_amount']),
                    # MatricItem(label="Avg Loss Amount",
                    #            value=assets[idx]['avg_loss_amount']),
                    # MatricItem(label="Avg Win %",
                    #            value=assets[idx]['avg_win_percent']),
                    # MatricItem(label="Avg Loss %",
                    #            value=assets[idx]['avg_loss_percent']),

                ]
            )
        ]
    )

@callback(
    Output("strategy-metrics", "children"),
    [
        Input(STRATEGY_STORE_MAPPINGS.metrics.id, 'data'),
    ],
    # prevent_initial_call=True,

)
def update_account_card_section(metrics: IStrategyMatrics):
    currency = "£"
    if metrics is None:
        return  None
    return  [
                    MatricItem(label="Win Rate", value=f"{round(metrics['win_rate'], 2)*100}%"),
                    MatricItem(label="Net PnL", value=f"{currency}{round(metrics['total_pnl'], 2)}"), 
                    MatricItem(label="Trades Open", value=metrics['total_open']),
                    MatricItem(label="Trades Closed", value=metrics['total_closed']),
                    MatricItem(label="Total Profits", value=f"{currency}{round(metrics['total_profit'], 2)}"),
                    MatricItem(label="Total Losses", value=f"{currency}{round(metrics['total_loss'], 2)}"),
                    MatricItem(label="Avg Win", value=f"{currency}{round(metrics['avg_win'], 2)}"),
                    MatricItem(label="Avg Loss", value=f"{currency}{round(metrics['avg_loss'], 2)}"),
    ]
 


def accountBalanceChart():
    return Tvlwc(
        id='account-balance-chart',
        seriesTypes=[SeriesType.Baseline],
        # seriesData=[],
        seriesOptions=[{
            'baseValue': {'type': 'price', 'price': 0},
            'topFillColor1': 'rgb(34, 197, 94)',
            'topFillColor2': 'rgba(34, 197, 94, 0)',
            'topLineColor': 'rgb(34, 197, 94)',
            'crosshairMarkerRadius': 4,
            'lineWidth': 2,
            'priceScaleId': 'left',
        }],
        width='100%',
        chartOptions={
            'rightPriceScale': {'visible': False},
            'leftPriceScale': {'visible': True, 'borderColor': '#be9e6b', },
            'timeScale': {'timeVisible': True, 'visible': True, 'secondsVisible': True},
            'grid': {'vertLines': {'visible': False}, 'horzLines': {'style': 0, 'color': '#be9e6b80'}},
            'layout': {'textColor': '#fffffa', 'background': {'type':  ColorType.Solid, 'color': '#0a0f1a'}},
            'localization': {
                    'timeFormatter': "(businessDayOrTimestamp) => {return Date(businessDayOrTimestamp);}",
                    # 'timeFormatter': "(businessDayOrTimestamp) => {return Date(businessDayOrTimestamp/1000);}",
                 
                #     'locale': 'en-US',
                #     'priceFormatter': "(function(price) { return '$' + price.toFixed(2); })"
                }

        }, 
        fullTimeScaleOptions={
            'visible': True,
		    'timeVisible': True,
            'secondsVisible': True,
        }
    )


@callback(
    [Output("account-balance-chart", "seriesData"),
     Output("account-balance-chart", 'seriesOptions')],
    [Input(STRATEGY_STORE_MAPPINGS.account.id, 'data')],
    [
         State(STRATEGY_STORE_MAPPINGS.metrics.id, 'data'),
         State("account-balance-chart", 'seriesData'),
         State("account-balance-chart",'seriesOptions'),
         State(STRATEGY_STORE_MAPPINGS.time.id, 'data')
    ],
    # prevent_initial_call=True,
    suppress_callback_exceptions=True
)
def update_account_balance_chart(account, metrics, balance_series, chart_options, time):
    updateBaseline = False

    # Check if the balance series is empty and populate the starting balance
    if balance_series == None:
        balance_series = [[]]
        if metrics is None:
            return no_update, no_update
        balance_series[0].append(
            {"time": metrics["start_date"], "value": metrics["starting_cash"]})
        chart_options[0]["baseValue"]["price"] = metrics["starting_cash"]
        updateBaseline = True

    # Check if the balance has changed
    if balance_series[0][-1]["value"] == account["equity"]:
        return  balance_series if updateBaseline else no_update, chart_options if updateBaseline else no_update
    
    # Add a new data point to the series
    new_datapoint = [{"time": time, "value": account["equity"]}]
    # if balance_series[0][-1]["time"]  < time:
    #     balance_series[0].extend(new_datapoint)
    balance_series[0].extend(new_datapoint)

    return balance_series, chart_options if updateBaseline else no_update



def assetAllocationChart():
    return dcc.Graph(id="asset-allocation-chart", className="bg-primary-light")

@callback(
    [Output("asset-allocation-chart", "figure")],
    [Input(STRATEGY_STORE_MAPPINGS.account.id, 'data'),
     Input(STRATEGY_STORE_MAPPINGS.positions.id, 'data')],
    # prevent_initial_call=True,
    suppress_callback_exceptions=True
)
def update_asset_allocation_chart(account: IAccount, positions: IPosition):
        names = ["Portfolio", "Cash"]
        parents = ["", "Portfolio"]
        values = [0, account["cash"]]

        # print(positions)
        # Populate the names, parents and values lists
        if positions is not None:
            for symbol, position in positions.items():
                names.append(symbol)
                if position is None:
                    continue
                parents.append("Portfolio")
                values.append(position["market_value"])
        
        asset_allocation = go.Figure(go.Sunburst(
            labels=names,
            parents=parents,
            values=values,

        ), layout=go.Layout(
            plot_bgcolor='white',
            paper_bgcolor='#0a0f1a',
            font={"color": 'white'},
        ))
        return [asset_allocation]

account_porfolio_chart_section = html.Div([
    PlaceholderChart(title="Portfolio Balance Chart",
                     children=accountBalanceChart()),
    PlaceholderChart(title="Portfolio Distribution",
                     children=assetAllocationChart()),


], className="mt-8 grid grid-cols-1 lg:grid-cols-2 gap-6")


insight_details_section = html.Div([
    # InsightDetails(insight)
], className="mt-8")

recent_trades_section = html.Div([
    html.H2("Recent Trades", className="text-2xl font-bold mb-4 text-accent"),
    tradeTable("dashboard"),
], className="mt-8")

layout = html.Div([
    subnav(),
    html.Div(
        className="flex-grow container mx-auto py-8 text-white",
        children=[
            account_card_section,
            strategy_metrics_section,
            account_porfolio_chart_section,
            insight_details_section,
            recent_trades_section
        ])
])



import os
from dotenv import load_dotenv

from OlympusTrader.strategy.sharedmemory import SharedStrategyManager

def get_shared_strategy_manager():
    """Get the shared strategy manager instance"""
    try:
        load_dotenv()
        assert os.getenv('SSM_HOST') and os.getenv(
            'SSM_PASSWORD'), "Please set SSM_HOST and SSM_PASSWORD environment variables"

        # print("Connecting to Shared Strategy Manager...")

        manager = SharedStrategyManager(address=(
            os.getenv("SSM_HOST"), 50000), authkey=os.getenv("SSM_PASSWORD").encode())

        manager.connect()
        
        return manager

    except Exception as e:
        return None


def get_strategiesPath():
    path = "."
    if os.getenv('STRATEGIES_PATH'):
        path = os.getenv('STRATEGIES_PATH')
    else:
        print("Please set STRATEGIES_PATH environment variable. Using default path '.'")

    return path



import numpy as np
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from OlympusTrader.strategy.base_strategy import BaseStrategy

from ..insight.insight import InsightState


class ITradingTools():
    STRATEGY = None

    def __init__(self, strategy):
        self.STRATEGY = strategy

    def dynamic_round(self, v: float, symbol: str) -> float:
        """Round float depending on log10 decimal places"""
        if "price_base" not in self.STRATEGY.assets[symbol] or self.STRATEGY.assets[symbol]["price_base"] == None:
            dynamic_precision = np.abs(
                int(np.log10(self.STRATEGY.assets[symbol]["min_price_increment"])))
            self.STRATEGY.UNIVERSE[symbol]["price_base"] = dynamic_precision

        return round(v, self.STRATEGY.UNIVERSE[symbol]["price_base"])
    # TODO: movethis into strategy tools
    def calculateTimeToLive(self, price: float, entry: float, ATR: float, additional: int = 2) -> int:
        """Calculate the time to live for a given price and entry based on the ATR"""
        return np.ceil((np.abs(price - entry)) / ATR)+additional
    
    def get_unrealized_pnl(self, symbol: str) -> float:
        """Calculate unrealized PnL for a given symbol"""
        if symbol not in self.STRATEGY.positions:
            return 0
        position = self.STRATEGY.positions[symbol]
        if position["side"] == "long":
            return (position["current_price"] - position["avg_entry_price"]) * position["qty"]
        else:
            return (position["avg_entry_price"] - position["current_price"]) * position["qty"]
    
    def get_all_unrealized_pnl(self) -> float:
        """Calculate unrealized PnL for a given symbol"""
        unrealized_pnl = 0
        for symbol in self.STRATEGY.positions:
            unrealized_pnl += self.get_unrealized_pnl(symbol)
        return unrealized_pnl
    
    def get_filled_insights(self) -> list:
        """Get all filled insights"""
        filled_insights = []
        for symbol in self.STRATEGY.insights:
           for insight in self.STRATEGY.insights[symbol]:
               if insight["status"] == InsightState.FILLED:
                   filled_insights.append(insight)
        return filled_insights

def dynamic_round(v: float) -> float:
    """Round float depending on log10 decimal places"""
    dynamic_precision = np.abs(int(np.log10(v)))+2
    return round(v, dynamic_precision)


if __name__ == '__main__':
    # tools = TradingTools()
    print(dynamic_round(0.0045456171111))  # 0.0045456171111
    print(dynamic_round(0.000045456171111))  # 0.0045456171111
    print(dynamic_round(18.82666666))  # 18.83
    print(dynamic_round(12.2222222))  # 12.22
    print(dynamic_round(122222222))  # 12.22
    print(dynamic_round(0.291))  # 0.25
    print(dynamic_round(0.000000001))  # 0.0
    print(dynamic_round(5505999.68))



# Strategy
from OlympusTrader.strategy.strategy import Strategy

# Account
from OlympusTrader.broker.interfaces import IAsset, IAccount, IPosition

# Insight
from OlympusTrader.insight import Insight, StrategyTypes, InsightState, IStrategyMode, StrategyDependantConfirmation
from OlympusTrader.insight.executors import BaseExecutor
from OlympusTrader.alpha import BaseAlpha

# Brokers
from OlympusTrader.broker.alpaca_broker import AlpacaBroker
from OlympusTrader.broker.ccxt_broker import CCXTBroker
from OlympusTrader.broker.paper_broker import PaperBroker
from OlympusTrader.broker.interfaces import IOrderSide

# Utiles and Tools
from OlympusTrader.utils.tools import ITradingTools, dynamic_round
from OlympusTrader.utils.timeframe import ITimeFrame, ITimeFrameUnit


from ..base_executor import BaseExecutor
from ...insight import InsightState


class DefaultOnRejectExecutor(BaseExecutor):
    """
    ### Executor for Default On Reject
    This executor is used to delete the insight from the strategy when it is rejected.

    :param strategy (BaseStrategy): The strategy instance

    Author:
        @isaac-diaby
    """

    def __init__(self, strategy):
        super().__init__(strategy, InsightState.REJECTED, "1.0")

    def run(self, insight):
        # Set the default state of the insight
        try:
            del self.STRATEGY.insights[insight.INSIGHT_ID]
            return self.returnResults(False, True, "Insight Rejected Successfully.")
        except Exception as e:
            return self.returnResults(False, False, f"Error deleting insight: {e}")



from dash import dcc, html


def subnav():
    return html.Nav(
    className="bg-primary-foreground border-b border-accent py-2",
    children=[
        html.Div(
            className="container mx-auto",
            children=[
                html.Ul(
                    className="flex space-x-4",
                    children=[
                        html.Li(
                            children=[
                                dcc.Link(
                                    "Dashboard",
                                    href="/",
                                    className="text-accent hover:text-white transition-colors"
                                )
                            ]
                        ),
                        html.Li(
                            children=[
                                dcc.Link(
                                    "Assets",
                                    href="/assets",
                                    className="text-accent hover:text-white transition-colors"
                                )
                            ]
                        ),
                        html.Li(
                            children=[
                                dcc.Link(
                                    "Settings",
                                    href="/settings",
                                    className="text-accent hover:text-white transition-colors"
                                )
                            ]
                        )
                    ]
                )
            ]
        )
    ]
)


import pandas as pd
from datetime import datetime
import pytz
import numpy as np

# TODO: we would want to make this function more performant by taking in candles that are already in the correct format and just convert the ones that are not
def history_to_trading_view_format(data: pd.DataFrame):
    """
    Convert the history dataframe to a trading view format
    """

    # IF the data is empty return an empty list
    if data.empty:
        return []
    
    history = data[["open", "high", "low", "close", "volume"]].copy()
    # if we can not pass the TA indicators values to the trading view chart
    # history = data.copy()
    history.reset_index(inplace=True)
    history.dropna(inplace=True)
    history.drop(columns=["symbol"], inplace=True)

    if "date" in history.columns:
        history.rename(columns={"date": "time"}, inplace=True)
    elif "timestamp" in history.columns:
        history.rename(columns={"timestamp": "time"}, inplace=True)
    else:
        history.rename(columns={"level_1": "time"}, inplace=True)



    # DEBUG: What the  date time looks like
    # print(history.tail(5))

    epoch = datetime(1970, 1, 1, tzinfo=pytz.UTC)
    history["time"] = history["time"].apply(lambda x: int((x - epoch).total_seconds()))

    # Sort the DataFrame by the timestamp column
    history.sort_values(by="time", inplace=True)

    return history.to_dict(orient='records')


def genMockDataFrame(days,startPrice,colName,startDate,seed=None): 
   
    periods = days*24
    np.random.seed(seed)
    steps = np.random.normal(loc=0, scale=0.0018, size=periods)
    steps[0]=0
    P = startPrice+np.cumsum(steps)
    P = [round(i,4) for i in P]

    fxDF = pd.DataFrame({ 
        # 'ticker':np.repeat( [colName], periods ),
        'time': np.tile( pd.date_range(startDate, periods=periods, freq='h'), 1 ),
        'price':(P)})
    fxDF.index = pd.to_datetime(fxDF.time)
    fxDF = fxDF.price.resample('D').ohlc()
    vol = np.random.randint(100,5000,days)
    fxDF['volume'] = vol
    return fxDF


import numpy as np
from ..base_executor import BaseExecutor
from ...insight import InsightState

from ....utils.tools import dynamic_round

class DynamicQuantityToRiskExecutor(BaseExecutor):
    """
    ### Executor for Dynamic Quantity to Risk
    This executor uses risks a percentage of the account balance to determine the quantity to trade.

    :param strategy (BaseStrategy): The strategy instance
    :param maximum_costbasis (float): The maximum cost basis for the trade. Default is 200,000.
    :param minimum_costbasis (float): The minimum cost basis for the trade. Default is 1,000.

    #### How it works:
    When a new insight is generated, the executor calculates the quantity to trade based on the percentage of the account balance which is a percentage of insight.confidence * RISK.

    Note: limit price and stop loss levels must be set in the insight before using this executor.

    Author:
        @isaac-diaby
    """
    maximum_costbasis: float
    minimum_costbasis: float
    def __init__(self, strategy, maximum_costbasis: float = 200_000.0, minimum_costbasis: float = 1_000.0):
        super().__init__(strategy, InsightState.NEW, "1.1")
        if self.STRATEGY.broker.supportedFeatures.maxOrderValue is not None:
            self.maximum_costbasis = min(self.STRATEGY.broker.supportedFeatures.maxOrderValue, maximum_costbasis)
        else:
            self.maximum_costbasis = maximum_costbasis
        self.minimum_costbasis = minimum_costbasis

        assert self.maximum_costbasis > self.minimum_costbasis, "Maximum cost basis must be greater than or equal to the minimum cost basis."

    def run(self, insight):
        if insight.limit_price is None or not insight.SL or np.isnan(insight.SL):
            return self.returnResults(False, False, "Insight does not have limit price or stop loss levels set.")
        
        # if the user has set the quantity, then we don't need to calculate it
        if insight.quantity is not None:
            return self.returnResults(True, True, "Quantity already set")

        # Calculate the quantity to trade
        try:
            #Add a Check if margin is enabled for this asset. If margin is enabled, then use buying_power to calculate the quantity.
            buying_power = self.STRATEGY.account.buying_power if self.STRATEGY.assets[insight.symbol]['marginable'] else self.STRATEGY.account.cash
            
            # Account size  to place order
            working_capital = buying_power * insight.confidence

            if working_capital < self.minimum_costbasis:
                response = self.returnResults(False, True, f"Working capital is less than the minimum cost basis: {self.minimum_costbasis} : Suggested: {working_capital}")
                self.changeState(
                    insight, InsightState.REJECTED, response.message)
                return response
            
            # Check if the working capital is greater than the maximum cost basis
            if working_capital > self.maximum_costbasis:
                working_capital = self.maximum_costbasis

            # Calculate the risk per share (PIP value)
            account_size_at_risk = working_capital * self.STRATEGY.execution_risk
            riskPerShare = abs(insight.limit_price - insight.SL)
        
            # Calculate the maximum quantity that can be bought
            maximun_can_buy = working_capital/insight.limit_price

            # Calculate the quantity to trade
            size_should_buy = account_size_at_risk/riskPerShare

            # Check if the quantity is greater than the maximum quantity that can be bought
            size_should_buy = min(size_should_buy, maximun_can_buy)

            if size_should_buy > 1:
                # Round down the quantity to the nearest whole number
                size_should_buy = np.floor(size_should_buy)
            else:
                #check if the asset is fractional or not
                if self.STRATEGY.assets[insight.symbol]['fractionable']:
                    # Although the quantity is less than 1, we can still buy a fraction of a share
                    # Round the quantity to the nearest 2 decimal places
                    # FIXME: This should always round down but right now it can round up. another function should be created to only round down dynamically to the closest minimum order size precision
                    size_should_buy = dynamic_round(size_should_buy)
                else:
                    response = self.returnResults(False, True, f"Asset is not fractionable: Suggested: {size_should_buy}")
                    self.changeState(
                        insight, InsightState.REJECTED, response.message)
                    return response

            # Check if the quantity is greater than the minimum order size
            minimum_order_size = self.STRATEGY.assets[insight.symbol]['min_order_size']

            # Check to see if we need to convert the to number of contracts
            if insight.uses_contract_size:
                size_should_buy = round(size_should_buy / self.STRATEGY.assets[insight.symbol]["contract_size"], 2)
                
            if size_should_buy < minimum_order_size:
                response = self.returnResults(False, True, f"Quantity is less than the minimum order size: {minimum_order_size} : Suggested: {size_should_buy}")
                self.changeState(
                    insight, InsightState.REJECTED, response.message)
                return response
            
            # Cap to max order size if it is set
            if self.STRATEGY.assets[insight.symbol].get('max_order_size', None) is not None:
                size_should_buy = min(size_should_buy, self.STRATEGY.assets[insight.symbol]['max_order_size'])


            # Update the quantity in the insight
            self.STRATEGY.insights[insight.INSIGHT_ID].update_quantity(
                size_should_buy)

            return self.returnResults(True, True, f"Quantity set to {size_should_buy}")
        except Exception as e:
            return self.returnResults(False, False, f"Error calculating quantity: {e}")



from datetime import datetime, timedelta
from enum import Enum


"""FROM ALPACA TIMEFRAME.PY but modified to be used for OlympusTrader framework"""


class ITimeFrameUnit(str, Enum):
    """Quantity of time used as unit"""

    Minute: str = "Min"
    Hour: str = "Hour"
    Day: str = "Day"
    Week: str = "Week"
    Month: str = "Month"


class ITimeFrame:
    """A time interval specified in multiples of defined units (minute, day, etc)

    Attributes:
        amount_value (int): The number of multiples of the ITimeFrameUnit interval
        unit_value (ITimeFrameUnit): The base unit of time interval that is used to measure the TimeFrame

    Raises:
        ValueError: Raised if the amount_value and unit_value are not consistent with each other
    """

    amount_value: int
    unit_value: ITimeFrameUnit

    def __init__(self, amount, unit) -> None:
        self.validate_timeframe(amount, unit)
        self.amount_value = amount
        self.unit_value = unit

    @property
    def amount(self) -> int:
        """Returns the amount_value field

        Returns:
            int: amount_value field
        """
        return self.amount_value

    @property
    def unit(self) -> ITimeFrameUnit:
        """Returns the ITimeFrameUnit field value of this TimeFrame object

        Returns:
            ITimeFrameUnit: unit_value field
        """
        return self.unit_value

    @property
    def value(self) -> str:
        """Returns a string representation of this TimeFrame object for API consumption

        Returns:
            str: string representation of this timeframe
        """
        return f"{self.amount}{self.unit.value}"

    @staticmethod
    def validate_timeframe(amount: int, unit: ITimeFrameUnit):
        """Validates the amount value against the ITimeFrameUnit value for consistency

        Args:
            amount (int): The number of multiples of unit
            unit (ITimeFrameUnit): The base unit of time interval the TimeFrame is measured by

        Raises:
            ValueError: Raised if the values of amount and unit are not consistent with each other
        """
        if amount <= 0:
            raise ValueError("Amount must be a positive integer value.")

        if unit == ITimeFrameUnit.Minute and amount > 59:
            raise ValueError(
                "Second or Minute units can only be "
                + "used with amounts between 1-59."
            )

        if unit == ITimeFrameUnit.Hour and amount > 23:
            raise ValueError("Hour units can only be used with amounts 1-23")

        if unit in (ITimeFrameUnit.Day, ITimeFrameUnit.Week) and amount != 1:
            raise ValueError(
                "Day and Week units can only be used with amount 1")

        if unit == ITimeFrameUnit.Month and amount not in (1, 2, 3, 6, 12):
            raise ValueError(
                "Month units can only be used with amount 1, 2, 3, 6 and 12"
            )
        return True
    
    def is_time_increment(self, time: datetime) -> bool:
        """ return true if the current date time is in the frequency of the time of X min increments
        Args:
            time (datetime): current time
            resolution (TimeFrame): TimeFrame object
            frequency (int, optional): frequency of the time. Defaults to 5.

        Returns:
            bool: True if the current date time is in the frequency of the time of X Time Unit increments

        Note: This will just match the Time Unit, not the exact time of the day. so will pass if the time is 12:00:00 or 12:00:59 same with dates and months
        Note: if your using this to execute a trade, you should check if the market is open first before executing a trade
        """
        match self.unit_value:
            case ITimeFrameUnit.Minute:
                if time.minute % self.amount_value == 0:
                    return True
                else:
                    return False
            case ITimeFrameUnit.Hour:

                if time.hour % self.amount_value == 0:
                    return True
                else:
                    return False
            case ITimeFrameUnit.Day:
                if time.day % self.amount_value == 0:
                    return True
                else:
                    return False
            case ITimeFrameUnit.Week:
                if time.week % self.amount_value == 0:
                    return True
                else:
                    return False
            case ITimeFrameUnit.Month:
                if time.month % self.amount_value == 0:
                    return True
                else:
                    return False
            case _:
                print("resolution Error: ITimeFrameUnit not implemented")
                return False
            
    def add_time_increment(self, time: datetime, periods: int) -> datetime:
        """ 
        Add the time increment to the current time
        """
        match self.unit_value:
            case ITimeFrameUnit.Minute:
                return time + timedelta(minutes=self.amount_value*periods, seconds=-time.second, microseconds=-time.microsecond)
            case ITimeFrameUnit.Hour:
                return time + timedelta(hours=self.amount_value*periods, minutes=-time.minute, seconds=-time.second, microseconds=-time.microsecond)
            case ITimeFrameUnit.Day:
                return time + timedelta(days=self.amount_value*periods, hours=-time.hour, minutes=-time.minute, seconds=-time.second, microseconds=-time.microsecond)
            case ITimeFrameUnit.Week:
                return time + timedelta(weeks=self.amount_value*periods, days=-time.day, hours=-time.hour, minutes=-time.minute, seconds=-time.second, microseconds=-time.microsecond)
            case ITimeFrameUnit.Month:
                return time + timedelta(months=self.amount_value*periods, days=-time.day, hours=-time.hour, minutes=-time.minute, seconds=-time.second, microseconds=-time.microsecond)
            case _:
                print("resolution Error: ITimeFrameUnit not implemented")
                return False
            
    def get_time_increment(self, time: datetime) -> datetime:
        """ 
        Get the current time frame 
        """
        match self.unit_value:
            case ITimeFrameUnit.Minute:
                return time - timedelta(minutes=time.minute % self.amount_value, seconds=time.second, microseconds=time.microsecond)
            case ITimeFrameUnit.Hour:
                return time - timedelta(hours=time.hour % self.amount_value, minutes=time.minute, seconds=time.second, microseconds=time.microsecond)
            case ITimeFrameUnit.Day:
                return time - timedelta(days=time.day % self.amount_value, hours=time.hour, minutes=time.minute, seconds=time.second, microseconds=time.microsecond)
            case ITimeFrameUnit.Week:
                return time - timedelta(weeks=time.week % self.amount_value, days=time.day, hours=time.hour, minutes=time.minute, seconds=time.second, microseconds=time.microsecond)
            case ITimeFrameUnit.Month:
                return time - timedelta(months=time.month % self.amount_value, days=time.day, hours=time.hour, minutes=time.minute, seconds=time.second, microseconds=time.microsecond)
            case _:
                print("resolution Error: ITimeFrameUnit not implemented")
                return False
    def get_next_time_increment(self, time: datetime) -> datetime:
        """ 
        Get the next time frame increment
        """
        match self.unit_value:
            case ITimeFrameUnit.Minute:
                return time + timedelta(minutes=self.amount_value - time.minute % self.amount_value, seconds=-time.second, microseconds=-time.microsecond)
            case ITimeFrameUnit.Hour:
                return time + timedelta(hours=self.amount_value - time.hour % self.amount_value, minutes=-time.minute, seconds=-time.second, microseconds=-time.microsecond)
            case ITimeFrameUnit.Day:
                return time + timedelta(days=self.amount_value - time.day % self.amount_value, hours=-time.hour, minutes=-time.minute, seconds=-time.second, microseconds=-time.microsecond)
            case ITimeFrameUnit.Week:
                return time + timedelta(weeks=self.amount_value - time.week % self.amount_value, days=-time.day, hours=-time.hour, minutes=-time.minute, seconds=-time.second, microseconds=-time.microsecond)
            case ITimeFrameUnit.Month:
                return time + timedelta(months=self.amount_value - time.month % self.amount_value, days=-time.day, hours=-time.hour, minutes=-time.minute, seconds=-time.second, microseconds=-time.microsecond)
            case _:
                print("resolution Error: ITimeFrameUnit not implemented")
                return False
    def __int__(self):
        match self.unit_value:
            case ITimeFrameUnit.Minute:
                return self.amount_value
            case ITimeFrameUnit.Hour:
                return self.amount_value * 60
            case ITimeFrameUnit.Day:
                return self.amount_value * (60 * 24)
            case ITimeFrameUnit.Week:
                return self.amount_value * (60 * 24 * 7)
            case ITimeFrameUnit.Month:
                return self.amount_value * (60 * 24 * 7 * 30)
            case _:
                return self.amount_value
    def __str__(self):
        return self.value






import abc
from typing import TYPE_CHECKING, override
from pandas import DataFrame

from ...broker.interfaces import IQuote

from ..insight import Insight, InsightState


def get_BaseStrategy():
    from ...strategy.base_strategy import BaseStrategy
    return BaseStrategy


if TYPE_CHECKING:
    from ...strategy.base_strategy import BaseStrategy


class ExecutorResults():
    """
    ### Executor Results
    This class is used to store the results of the executor.
    """
    passed: bool
    """Boolean value indicating if the executor passed or not."""
    success: bool
    """Boolean value indicating if the executor was ran successful or not."""
    executor: str
    """Name of the executor."""
    message: str
    """Message indicating the result of the executor."""

    def __init__(self, passed: bool, success: bool = True, message: str = None, executor: str = None):
        self.passed = passed
        if passed:
            # If the executor passed, then it was successfully ran
            success = True
        self.success = success
        if self.success == False:
            # If the executor was not successfully ran, then it did not pass
            self.passed = False
        self.message = message
        self.executor = executor


class BaseExecutor(abc.ABC):
    """
    ### Abstract class for BaseExecutor implementations.
    Used to manage orders based on insights.
    """
    NAME: str
    VERSION: str
    """Version of the executor model."""
    STRATEGY: get_BaseStrategy
    """Reference to the strategy instance."""

    state: InsightState
    """Reference to the working state of the executor"""

    @abc.abstractmethod
    def __init__(self, strategy: get_BaseStrategy, state: InsightState, version: float = "1.0") -> None:
        self.NAME = self.__class__.__name__
        self.VERSION = version
        # Reference to the strategy instance
        self.STRATEGY = strategy
        self.state = state

    @override
    @abc.abstractmethod
    def run(self, insight: Insight) -> ExecutorResults:
        """Run the executor."""
        pass

    def returnResults(self, passed: bool, success: bool = True, message: str = None) -> ExecutorResults:
        return ExecutorResults(passed, success, message, self.NAME)

    # @property
    # def insight(self, insight: Insight):
    #     return self.STRATEGY.insights[insight.INSIGHT_ID]

    def changeState(self, insight: Insight, state: InsightState, message: str = None) -> None:
        self.STRATEGY.insights[insight.INSIGHT_ID].updateState(state, message)

    def get_history(self, symbol: str) -> DataFrame:
        return self.STRATEGY.history[symbol].loc[symbol]

    def get_latest_bar(self, symbol: str) -> DataFrame:
        return self.get_history(symbol).iloc[-1]

    def get_previos_bar(self, symbol: str) -> DataFrame:
        return self.get_history(symbol).iloc[-2]

    def get_latest_quote(self, insight: Insight) -> IQuote:
        return self.STRATEGY.broker.get_latest_quote(insight.ASSET)

    def _override_state(self, state: InsightState) -> None:
        assert isinstance(
            state, InsightState), "State must be an instance of InsightState"
        self.state = state



from ..base_executor import BaseExecutor
from ...insight import InsightState


class RejectExpiredInsightExecutor(BaseExecutor):
    """
    ### Executor for Rejecting Expired Insights
    This executor is used to reject insights that have expired.

   :param strategy (BaseStrategy): The strategy instance
   
    Author:
        @isaac-diaby
    """

    def __init__(self, strategy):
        super().__init__(strategy, InsightState.NEW, "1.0")

    def run(self, insight):
        # Check if the insight has expired
        try:
            hasExpired = self.STRATEGY.insights[insight.INSIGHT_ID].hasExpired(
                True)
            if hasExpired == None:
                response = self.returnResults(
                    True, True, "Insight state is not applicable for expiration check.")
                return response
            if hasExpired == True:
                response = self.returnResults(
                    False, True, "Insight has expired. Rejecting insight.")
                return response

            return self.returnResults(True, True)
        except Exception as e:
            return self.returnResults(False, False, f"Error rejecting expired insight: {e}")



from ..base_executor import BaseExecutor
from ...insight import InsightState


class DefaultOnCancelledExecutor(BaseExecutor):
    """
    ### Executor for Default On Cancelled
    This executor is used to delete the insight from the strategy when it is cancelled.

    :param strategy (BaseStrategy): The strategy instance

    Author:
        @isaac-diaby
    """

    def __init__(self, strategy):
        super().__init__(strategy, InsightState.CANCELED, "1.0")

    def run(self, insight):
        # Set the default state of the insight
        try:
            for i, order in self.STRATEGY.orders.items():
                if i == insight.order_id:
                    # Check if the insight is already filled
                    if (self.STRATEGY.insights[insight.INSIGHT_ID].state == InsightState.FILLED):
                        return self.returnResults(True, True, "Insight already filled. Not deleting.")

            del self.STRATEGY.insights[insight.INSIGHT_ID]
            return self.returnResults(True, True, "Insight Cancelled Successfully.")
        except Exception as e:
            return self.returnResults(False, False, f"Error deleting insight: {e}")
